# Copyright 2010-2011 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

__metaclass__ = type

__all__ = [
    "DeferredProperty",
    "DeferredBase",
    "implement_from_list",
    ]

from twisted.internet.defer import (
    Deferred,
    DeferredList,
    )


def get_deferred_obj(obj):
    try:
        return obj.__deferred_object__
    except AttributeError:
        deferred_obj = DeferredObject(obj)
        return obj.__dict__.setdefault("__deferred_object__", deferred_obj)


def get_deferred_cls(cls):
    if "__deferred_class__" in cls.__dict__:
        # Can't use attribute access here, otherwise subclassing won't work.
        return cls.__dict__["__deferred_class__"]
    else:
        cls.__deferred_class__ = DeferredClass(cls)
        return cls.__deferred_class__


class DeferredObject:
    """Persistent deferred information about an object."""

    # For get_deferred_obj(), an DeferredObject is its own deferred_obj.
    __deferred_object__ = property(lambda self: self)

    def __init__(self, obj):
        self.obj = obj
        self.deferred_cls = get_deferred_cls(type(obj))

        self.deferreds = deferreds = {}
        for prop in self.deferred_cls.properties:
            deferreds[prop] = Deferred()

        self.deferred_list = DeferredList(deferreds.values())
        self.deferred_list.chainDeferred(obj)


class DeferredClass:
    """Persistent deferred information about a class."""

    def __init__(self, cls):
        self.cls = cls

        pairs = []
        for attr in dir(cls):
            prop = getattr(cls, attr, None)
            if isinstance(prop, DeferredProperty):
                pairs.append((attr, prop))

        pairs.sort()

        self.properties = tuple(pair[1] for pair in pairs)
        self.attributes = dict(pairs)


class DeferredProperty:

    def __get__(self, obj, cls=None):
        if obj is None:
            return self
        deferred_obj = get_deferred_obj(obj)
        if cls is None:
            cls = deferred_obj.deferred_cls.cls
        deferred = deferred_obj.deferreds[self]
        if deferred.called:
            return deferred.result
        else:
            return None

    def __set__(self, obj, value):
        deferred_obj = get_deferred_obj(obj)
        deferred = deferred_obj.deferreds[self]
        if value is not None and not deferred.called:
            deferred.callback(value)


class DeferredBase(Deferred, object):

    pass


def implement_from_list(class_name, attributes=[], superclass=DeferredBase):
    """Return a class with the given attributes as deferred properties.

    :param attributes: A list of attributes to use when generating the class.
    :param superclass: The superclass of the class to be generated.
    """
    class_dict = {}
    for attribute in attributes:
        class_dict[attribute] = DeferredProperty()

    if not isinstance(superclass, tuple):
        superclass = (superclass,)
    return type(class_name, superclass, class_dict)
